from .base import ScFileException
from .basic import (
    FileBasicError,
    FileNotFound,
    FileSignatureInvalid,
    FileSuffixUnsupported,
)
from .decode import FileDecodingError, FileParsingError, FileUnsupportedError
from .mcsa import (
    McsaCountsLimit,
    McsaDecodingError,
    McsaUnknownLinkCount,
    McsaUnsupportedVersion,
)
from .ol import OlDecodingError, OlUnknownFourcc


__all__ = (
    "ScFileException",
    "FileBasicError",
    "FileNotFound",
    "FileSuffixUnsupported",
    "FileSignatureInvalid",
    "FileDecodingError",
    "FileParsingError",
    "FileUnsupportedError",
    "McsaCountsLimit",
    "McsaDecodingError",
    "McsaUnknownLinkCount",
    "McsaUnsupportedVersion",
    "OlDecodingError",
    "OlUnknownFourcc",
)



from scfile.consts import Magic
from scfile.file.base import FileEncoder
from scfile.file.data import ImageData


class PngEncoder(FileEncoder[ImageData]):
    @property
    def magic(self):
        return Magic.PNG

    def serialize(self):
        self.b.write(self.data.image)



from dataclasses import dataclass, field
from typing import List

from scfile.utils.model.mesh import Mesh
from scfile.utils.model.skeleton import Skeleton


@dataclass
class Flags:
    skeleton: bool = False
    texture: bool = False
    normals: bool = False


@dataclass
class Scale:
    position: float = 1.0
    texture: float = 1.0
    normals: float = 1.0
    weight: float = 1.0


@dataclass
class Model:
    flags: Flags = field(default_factory=Flags)
    scale: Scale = field(default_factory=Scale)
    meshes: List[Mesh] = field(default_factory=list)
    skeleton: Skeleton = field(default_factory=Skeleton)

    @property
    def total_vertices(self):
        return sum(mesh.count.vertices for mesh in self.meshes)

    @property
    def total_polygons(self):
        return sum(mesh.count.polygons for mesh in self.meshes)

    def ensure_unique_names(self):
        """Updates meshes names, excluding repetitions."""
        seen_names: set[str] = set()

        for mesh in self.meshes:
            name = mesh.name or "noname"

            base_name, count = name, 2
            unique_name = f"{base_name}"

            while unique_name in seen_names:
                unique_name = f"{base_name}_{count}"
                count += 1

            mesh.name = unique_name
            seen_names.add(unique_name)

    def convert_polygons_to_global(self, start: int = 0):
        """Updates meshes global_polygons indexes."""
        offset = start

        for mesh in self.meshes:
            mesh.global_polygons = [polygon >> offset for polygon in mesh.polygons]
            offset += mesh.count.vertices



from scfile.consts import OlString
from scfile.enums import StructFormat as F

from .binary import BinaryFileIO


class OlFileIO(BinaryFileIO):
    def readsizes(self, mipmap_count: int) -> list[int]:
        """Read lz4 uncompressed or compressed sizes."""
        return [self.readb(F.U32) for _ in range(mipmap_count)]

    def readfourcc(self) -> bytes:
        """Read xor encoded FourCC."""
        # Read string and skip last 0x00 byte
        string = self.read(OlString.SIZE)[:-1]

        # Xor byte if it's is not null
        return bytes(byte ^ OlString.XOR for byte in string if byte != OlString.NULL)



from enum import IntEnum, auto


class Flag(IntEnum):
    SKELETON = 0
    TEXTURE = auto()
    NORMALS = auto()
    TANGENTS = auto()
    BITANGENTS = auto()
    COLORS = auto()


def to_named_dict(flags: dict[int, bool]):
    return {Flag(key).name: value for key, value in flags.items()}



from pathlib import Path

from scfile.utils.types import PathLike

from .base import ScFileException


class FileDecodingError(ScFileException):
    """Exception occurring when file decoding."""

    def __init__(self, path: PathLike):
        self.path = Path(path)

    @property
    def posix_path(self) -> str:
        return self.path.as_posix()


class FileParsingError(ScFileException):
    """Exception occurring due incorrect file parsing. (Or file incorrect itself)."""

    pass


class FileUnsupportedError(ScFileException):
    """Exception occurring intentionally due incorrect file parsing."""

    pass



from dataclasses import dataclass

from .base import FileData


@dataclass
class TextureData(FileData):
    width: int
    height: int
    mipmap_count: int
    linear_size: int
    fourcc: bytes
    image: bytes



from enum import StrEnum, auto


class FileSuffix(StrEnum):
    DAE = auto()
    DDS = auto()
    MCSA = auto()
    MCVD = auto()
    MIC = auto()
    MS3D = auto()
    OBJ = auto()
    OL = auto()
    PNG = auto()
    TXT = auto()


class FileMode(StrEnum):
    RB = READ = auto()
    WB = WRITE = auto()
    AB = APPEND = auto()
    PLUS = "+"


class ByteOrder(StrEnum):
    NATIVE = "@"
    STANDARD = "="
    LITTLE = "<"
    BIG = ">"
    NETWORK = "!"


class StructFormat(StrEnum):
    BOOL = "?"
    """boolean: `1 byte` [False, True]"""

    I8 = "b"
    """signed char: `BYTE` `1 byte` [-128, 127]"""
    I16 = "h"
    """signed short: `WORD` `2 bytes` [-32768, 32767]"""
    I32 = "i"
    """signed int: `DWORD` `4 bytes` [-2147483648, 2147483647]"""

    U8 = "B"
    """unsigned char: `BYTE` `1 byte` [0, 255]"""
    U16 = "H"
    """unsigned short: `WORD` `2 bytes` [0, 65535]"""
    U32 = "I"
    """unsigned int: `DWORD` `4 bytes` [0, 4294967295]"""

    F16 = "e"
    """float: `half-precision` `2 bytes`"""
    F32 = "f"
    """double: `single-precision` `4 bytes`"""



import io
import struct
from typing import Any, Optional

from scfile.enums import ByteOrder
from scfile.enums import StructFormat as F


class StructBufferedIOBase(io.BufferedIOBase):
    def unpack(self, fmt: str) -> Any:
        """Unpack binary data from buffer."""
        size = struct.calcsize(str(fmt))
        data = self.read(size)
        return struct.unpack(fmt, data)

    def pack(self, fmt: str, *v: Any) -> bytes:
        """Pack values into a bytes."""
        return struct.pack(str(fmt), *v)


class BinaryIO(StructBufferedIOBase):
    order = ByteOrder.LITTLE
    """Binary byte order."""

    def readb(self, fmt: str, order: Optional[ByteOrder] = None) -> Any:
        """Unpack any binary data."""
        order = order or self.order
        return self.unpack(f"{order}{fmt}")[0]

    def reads(self, order: Optional[ByteOrder] = None) -> bytes:
        """Unpack length-prefixed string."""
        size = self.readb(F.U16, order)
        return self.unpack(f"{size}s")[0]

    def writeb(self, fmt: str, *v: Any, order: Optional[ByteOrder] = None) -> None:
        """Pack any binary data."""
        order = order or self.order
        data = self.pack(f"{order}{fmt}", *v)
        self.write(data)

    def writen(self, count: int = 1, size: int = 4) -> None:
        """Fill buffer with zeros by size (4 bytes)."""
        self.write(bytes(size * count))

    def writes(self, string: str) -> None:
        """Write utf-8 encoded string."""
        self.write(string.encode("utf-8", errors="replace"))



from abc import ABC, abstractmethod
from io import IOBase
from typing import Any, Optional, Type

from scfile.file.data import FileData


class BaseFile(ABC):
    @property
    @abstractmethod
    def buffer(self) -> IOBase:
        """IOBase buffer for reading or writing data."""
        pass

    @property
    @abstractmethod
    def data(self) -> Optional[FileData]:
        """Parsed file data."""
        pass

    def close(self):
        """Closes buffer handler."""
        self.buffer.close()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        self.close()



from .enums import FileSuffix


class Signature:
    """
    Big-Endian Source File Signature.
    Integer.
    """

    MCSA = 0x4D435341
    MIC = 0x894D4943
    OL = 0x0A9523FD


class Magic:
    """
    Big-Endian Output File Signature.
    List[Integer].
    """

    DDS = [0x44, 0x44, 0x53, 0x20]
    PNG = [0x89, 0x50, 0x4E, 0x47]


class Factor:
    """
    Integer range.
    """

    I8 = 0x7F
    U8 = 0xFF
    I16 = 0x7FFF
    U16 = 0xFFFF


class OlString:
    """
    Xor encoded zero-end string.
    """

    SIZE = 17
    """16 encoded bytes and last 0x00 byte."""
    XOR = ord("g")
    NULL = ord("G")


class McsaModel:
    """
    Mcsa model constants.
    """

    ROOT_BONE_ID = -1
    COUNT_LIMIT = 0x100000
    """There is no known limit. This for case when file was read incorrectly, so as not to overflow memory."""


class McsaSize:
    """
    Count of elements in each mcsa data structure.
    """

    POSITION = 4
    TEXTURE = 2
    NORMALS = 4
    POLYGONS = 3
    COLOR = 4


SUPPORTED_SUFFIXES = {FileSuffix.MCSA, FileSuffix.MCVD, FileSuffix.MIC, FileSuffix.OL}
"""Files suffixes that can be converted."""


CUBEMAP_FACES = 6
"""Dds cubemap faces count. (+x, -x, +y, -y, +z, -z)."""



import lz4.block

from scfile import exceptions as exc
from scfile.consts import Signature
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as F
from scfile.file.base import FileDecoder
from scfile.file.data import TextureData
from scfile.file.formats.dds import DdsEncoder
from scfile.io import OlFileIO

from .formats import SUPPORTED_FORMATS


class OlDecoder(FileDecoder[OlFileIO, TextureData]):
    def to_dds(self):
        return self.convert_to(DdsEncoder)

    @property
    def opener(self):
        return OlFileIO

    @property
    def order(self):
        return ByteOrder.BIG

    @property
    def signature(self):
        return Signature.OL

    def create_data(self):
        return TextureData(
            self.width, self.height, self.mipmap_count, self.linear_size, self.fourcc, self.image
        )

    def parse(self):
        self._parse_header()
        self._parse_fourcc()
        self._parse_sizes()
        self._parse_image()

    def _parse_header(self):
        self.width = self.f.readb(F.U32)
        self.height = self.f.readb(F.U32)
        self.mipmap_count = self.f.readb(F.U32)

    def _parse_fourcc(self):
        self.fourcc = self.f.readfourcc()

        if self.fourcc not in SUPPORTED_FORMATS:
            raise exc.OlUnknownFourcc(self.path, self.fourcc.decode())

        self.fourcc = b"ATI2" if self.fourcc == b"DXN_XY" else self.fourcc

    def _parse_sizes(self):
        self.uncompressed = self.f.readsizes(self.mipmap_count)
        self.compressed = self.f.readsizes(self.mipmap_count)

    def _decompress_mipmaps(self):
        self.mipmaps: list[bytes] = []

        for mipmap in range(self.mipmap_count):
            self.mipmaps.append(
                lz4.block.decompress(
                    self.f.read(self.compressed[mipmap]), self.uncompressed[mipmap]
                )
            )

    def _parse_image(self):
        self.texture_id = self.f.reads()

        self._decompress_mipmaps()

        self.image = b"".join(self.mipmaps)

    @property
    def linear_size(self) -> int:
        return self.uncompressed[0]



import io
from pathlib import Path

from .base import BinaryIO


class BinaryBytesIO(io.BytesIO, BinaryIO):
    pass


class BinaryFileIO(io.FileIO, BinaryIO):  # type: ignore
    @property
    def path(self) -> Path:
        return Path(str(self.name))



from scfile.consts import Signature
from scfile.file.base import FileDecoder
from scfile.file.data import ImageData
from scfile.file.formats.png import PngEncoder
from scfile.io import BinaryFileIO


class MicDecoder(FileDecoder[BinaryFileIO, ImageData]):
    def to_png(self):
        return self.convert_to(PngEncoder)

    @property
    def opener(self):
        return BinaryFileIO

    @property
    def signature(self):
        return Signature.MIC

    def create_data(self):
        return ImageData(self.image)

    def parse(self):
        self.image = self.f.read()



from pathlib import Path
from typing import Optional

from rich import print

from scfile.exceptions import ScFileException
from scfile.utils import convert


def convert_file(src: Path, dest: Optional[Path] = None):
    try:
        convert.auto(src, dest)

    except ScFileException as err:
        print(f"\n[b red]Error:[/] {err}")

    except Exception as err:
        print(f"\n[b red]Unknown Error:[/] '{src}' - {err}")


def convert_multiple_files(files: tuple[Path], dest: Optional[Path] = None):
    for src in files:
        convert_file(src, dest)



from pathlib import Path
from typing import Optional

import click
from rich import print

from .consts import Types
from .convert import convert_multiple_files


@click.command(no_args_is_help=True)
@click.argument("files", type=Types.FILES, nargs=-1)
@click.option(
    "-O",
    "--output",
    help="Optional path to output directory.",
    type=Types.OUTPUT,
    multiple=False,
    nargs=1,
)
def default(files: tuple[Path], output: Optional[Path] = None):
    convert_multiple_files(files, output)


def main():
    try:
        default()

    except KeyboardInterrupt:
        print("\n\n[b yellow]Operation aborted.[/]")



from .data import FileData, ImageData, ModelData, TextureData
from .formats import (
    DaeEncoder,
    DdsEncoder,
    McsaDecoder,
    MicDecoder,
    Ms3dAsciiEncoder,
    Ms3dBinEncoder,
    ObjEncoder,
    OlDecoder,
    PngEncoder,
)


__all__ = (
    "FileData",
    "ImageData",
    "ModelData",
    "TextureData",
    "DaeEncoder",
    "DdsEncoder",
    "McsaDecoder",
    "MicDecoder",
    "Ms3dAsciiEncoder",
    "Ms3dBinEncoder",
    "ObjEncoder",
    "OlDecoder",
    "PngEncoder",
)



from scfile.consts import McsaModel
from scfile.enums import StructFormat as F
from scfile.file.base import FileEncoder
from scfile.file.data import ModelData


class Ms3dBinEncoder(FileEncoder[ModelData]):
    def serialize(self):
        self.model = self.data.model
        self.meshes = self.data.model.meshes
        self.skeleton = self.data.model.skeleton

        self.model.ensure_unique_names()
        self.model.convert_polygons_to_global()
        self.skeleton.convert_to_local()

        self._add_header()
        self._add_vertices()
        self._add_triangles()
        self._add_groups()
        self._add_materials()
        self._add_joints()

    def _add_header(self):
        self.b.writes("MS3D000000")  # 10 bytes signature
        self.b.writeb(F.I32, 0x4)  # version

    def _add_vertices(self):
        self.b.writeb(F.U16, self.model.total_vertices)

        for mesh in self.meshes:
            for v in mesh.vertices:
                pos = v.position
                # i8 flags, f32 pos[3], i8 bone id, u8 reference count
                # TODO: reference count
                self.b.writeb(F.I8 + (F.F32 * 3) + F.I8 + F.U8, 0, pos.x, pos.y, pos.z, -1, 0xFF)

    def _add_triangles(self):
        self.b.writeb(F.U16, self.model.total_polygons)

        for index, mesh in enumerate(self.meshes):
            for p, gp in zip(mesh.polygons, mesh.global_polygons):
                v1 = mesh.vertices[p.a]
                v2 = mesh.vertices[p.b]
                v3 = mesh.vertices[p.c]

                # ! its awful but faster
                # u16 flags, u16 indices[3]
                # f32 normals[3][3], f32 textures u[3], f32 textures v[3]
                # u8 smoothing group, u8 group index
                self.b.writeb(
                    F.U16
                    + (F.U16 * 3)
                    + (F.F32 * 3)
                    + (F.F32 * 3)
                    + (F.F32 * 3)
                    + (F.F32 * 3)
                    + (F.F32 * 3)
                    + F.U8
                    + F.U8,
                    0,
                    gp.a,
                    gp.b,
                    gp.c,
                    v1.normals.x,
                    v1.normals.y,
                    v1.normals.z,
                    v2.normals.x,
                    v2.normals.y,
                    v2.normals.z,
                    v3.normals.x,
                    v3.normals.y,
                    v3.normals.z,
                    v1.texture.u,
                    v2.texture.u,
                    v3.texture.u,
                    v1.texture.v,
                    v2.texture.v,
                    v3.texture.v,
                    1,
                    index,
                )

    def _add_groups(self):
        self.b.writeb(F.U16, len(self.meshes))  # groups count

        offset = 0

        for index, mesh in enumerate(self.meshes):
            self.b.writeb(F.U8, 0)  # flags
            self.b.write(self._fixedlen(mesh.name))  # group name

            self.b.writeb(F.U16, mesh.count.polygons)  # triangles count

            for p in range(len(mesh.polygons)):
                self.b.writeb(F.U16, p + offset)  # triangles indexes

            self.b.writeb(F.I8, index)  # material index

            offset += len(mesh.polygons)

    def _add_materials(self):
        self.b.writeb(F.U16, len(self.meshes))  # materials count

        for mesh in self.meshes:
            self.b.write(self._fixedlen(mesh.material))  # material name
            self.b.writeb(F.F32 * 4, 0.2, 0.2, 0.2, 1.0)  # ambient rgba
            self.b.writeb(F.F32 * 4, 0.8, 0.8, 0.8, 1.0)  # diffuse rgba
            self.b.writeb(F.F32 * 4, 0.0, 0.0, 0.0, 1.0)  # specular rgba
            self.b.writeb(F.F32 * 4, 0.0, 0.0, 0.0, 1.0)  # emissive rgba
            self.b.writeb(F.F32, 0.0)  # shininess
            self.b.writeb(F.F32, 1.0)  # transparency
            self.b.writeb(F.I8, 1)  # mode
            self.b.writen(count=128, size=1)  # texture
            self.b.writen(count=128, size=1)  # alphamap

    def _add_joints(self):
        self.b.writeb(F.F32, 24)  # fps
        self.b.writeb(F.F32, 1)  # current frame
        self.b.writeb(F.F32, 30)  # total frames

        self.b.writeb(F.U16, len(self.skeleton.bones))

        for bone in self.skeleton.bones:
            self.b.writeb(F.U8, 0)  # flags
            self.b.write(self._fixedlen(bone.name))  # bone name

            parent = self.skeleton.bones[bone.parent_id].name
            parent_name = parent if bone.parent_id != McsaModel.ROOT_BONE_ID else ""

            self.b.write(self._fixedlen(parent_name))  # parent name

            rot = bone.rotation
            pos = bone.position
            self.b.writeb(F.F32 * 3, rot.x, rot.y, rot.z)  # rotation
            self.b.writeb(F.F32 * 3, pos.x, pos.y, pos.z)  # position

            self.b.writeb(F.U16, 0)  # count keyframes rotation
            self.b.writeb(F.U16, 0)  # count keyframes transition

    def _fixedlen(self, name: str) -> bytes:
        return name.encode("utf-8").ljust(32, b"\x00")



from .encoder import DdsEncoder
from .header import DDS, CubemapFlag, Header, HeaderFlag, PixelFormat, PixelFormatFlag


__all__ = (
    "DdsEncoder",
    "DDS",
    "CubemapFlag",
    "Header",
    "HeaderFlag",
    "PixelFormat",
    "PixelFormatFlag",
)



from typing import Any, Dict

import numpy as np
from numpy.typing import NDArray

from scfile.consts import Factor, McsaModel, McsaSize
from scfile.enums import StructFormat as F
from scfile.exceptions.mcsa import McsaCountsLimit

from .binary import BinaryFileIO


class McsaFileIO(BinaryFileIO):
    FLOATS_ROUND = 6

    def readstring(self) -> str:
        """Read length-prefixed utf-8 string."""
        return self.reads().decode("utf-8", errors="replace")

    def readcounts(self):
        """Read unsigned int and validates count limit."""
        counts = self.readb(F.U32)
        if counts > McsaModel.COUNT_LIMIT:
            raise McsaCountsLimit(self.path, counts)
        return counts

    def readarray(self, fmt: str, size: int, count: int):
        """Read repetitive data values."""
        data = self.unpack(f"{size*count}{fmt}")
        return np.array(data, dtype=np.dtype(fmt))

    def readvertex(self, fmt: str, factor: float, size: int, count: int, scale: float = 1.0):
        """Read vertex data and scale it to floats."""
        # Read array
        data = self.readarray(fmt=fmt, size=size, count=count)

        # Scale values to floats
        data = data * scale / (factor)

        # Round digits
        data = data.round(self.FLOATS_ROUND)

        return self._reshape(data, size)

    def readpolygons(self, count: int):
        """Read vertex indexes."""
        size = McsaSize.POLYGONS

        # Validate that indexes fits into U16 range. Otherwise use U32.
        indexes = count * size
        fmt = F.U16 if indexes < Factor.U16 else F.U32

        # Read array
        data = self.readarray(fmt=fmt, size=size, count=count)

        return self._reshape(data, size)

    def _boneids_to_global(self, data: Any, mesh_bones: Dict[int, int]):
        def apply_mesh_bones(x: int):
            return mesh_bones.get(x, McsaModel.ROOT_BONE_ID)

        vectorized = np.vectorize(apply_mesh_bones)

        return vectorized(data)

    def readlinkspacked(self, count: int, mesh_bones: Dict[int, int]):
        # Read array
        data = self.readarray(fmt=F.U8, size=4, count=count)

        # Reshape to vertex[ids[2], weights[2]]
        data = data.reshape(-1, 2, 2)

        # Convert mesh bone ids to global skeleton ids
        ids = self._boneids_to_global(data[..., 0], mesh_bones)

        # Scale weights to floats
        weights = data[..., 1] / Factor.U8

        # Round digits
        weights = weights.round(self.FLOATS_ROUND)

        return (ids.tolist(), weights.tolist())

    def _readlinksids(self, count: int, mesh_bones: Dict[int, int]):
        # Read array
        size = 4
        data = self.readarray(fmt=F.U8, size=size, count=count)

        # Convert mesh bone ids to global skeleton ids
        data = self._boneids_to_global(data, mesh_bones)

        return self._reshape(data, size)

    def _readlinksweights(self, count: int):
        # Read array
        size = 4
        data = self.readarray(fmt=F.U8, size=size, count=count)

        # Scale weights to floats
        data = data / Factor.U8

        # Round digits
        data = data.round(self.FLOATS_ROUND)

        return self._reshape(data, size)

    def readlinksplains(self, count: int, mesh_bones: Dict[int, int]):
        ids = self._readlinksids(count=count, mesh_bones=mesh_bones)
        weights = self._readlinksweights(count=count)
        return (ids, weights)

    def _reshape(self, data: NDArray[Any], size: int) -> list[list[Any]]:
        """Reshapes array to list of lists."""
        # tolist: no further manipulation is provided
        # and python works faster with native lists (for encoders)
        return data.reshape(-1, size).tolist()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, Optional, TypeVar

from scfile.enums import ByteOrder, FileMode
from scfile.enums import StructFormat as F
from scfile.exceptions import FileSignatureInvalid
from scfile.file.data import FileData
from scfile.io import BinaryFileIO
from scfile.utils.types import PathLike

from .encoder import FileEncoder
from .file import BaseFile


OPENER = TypeVar("OPENER", bound=BinaryFileIO)
DATA = TypeVar("DATA", bound=FileData)


class FileDecoder(BaseFile, Generic[OPENER, DATA], ABC):
    def __init__(self, path: PathLike):
        self._path = path

        self._file = self.opener(self.path, self.mode)
        self._file.order = self.order

        self._data: Optional[DATA] = None

    @property
    def buffer(self):
        return self._file

    @property
    def data(self):
        return self._data

    @property
    def f(self):
        """File buffer abbreviation."""
        return self._file

    @property
    def path(self) -> Path:
        """Pathlib object to an open file."""
        return Path(self._path)

    @property
    def order(self) -> ByteOrder:
        """File reading default byte order."""
        return ByteOrder.LITTLE

    @property
    def signature(self) -> Optional[int]:
        """Optional file signature (4 bytes)."""
        return None

    @property
    def mode(self) -> str:
        """File opener mode."""
        return FileMode.READ

    @property
    @abstractmethod
    def opener(self) -> type[OPENER]:
        """Opener (reader) class for file."""
        pass

    @abstractmethod
    def create_data(self) -> DATA:
        """Creates an object with filled parsed data."""
        pass

    @abstractmethod
    def parse(self) -> None:
        """Reading file and entering values into data."""
        pass

    def decode(self) -> DATA:
        """File decoding. Signature validation, parsing and data creation."""
        self.validate_signature()
        self.parse()
        self._data = self.create_data()
        self.seek_to_start()
        return self._data

    def convert_to(self, encoder: type[FileEncoder[DATA]]) -> FileEncoder[DATA]:
        """Converting file into an object of passed encoder. Encoder has an open buffer."""
        data = self.decode()
        enc = encoder(data)
        enc.encode()
        return enc

    def convert(self, encoder: type[FileEncoder[DATA]]) -> bytes:
        """Converting file into content bytes of passed encoder."""
        enc = self.convert_to(encoder)
        content = enc.content
        enc.close()
        return content

    def read_signature(self) -> int:
        """Reading big-endian u32 (4 bytes) value."""
        return self.f.readb(F.U32, ByteOrder.BIG)

    def validate_signature(self) -> None:
        """Validates 4 bytes signature and throwing error in case of mismatch."""
        if self.signature:
            readed = self.read_signature()

            if readed != self.signature:
                raise FileSignatureInvalid(self.path, readed, self.signature)

    def seek_to_start(self) -> None:
        """Sets buffer position to start."""
        self.f.seek(0)



from .file import (
    DaeEncoder,
    DdsEncoder,
    McsaDecoder,
    MicDecoder,
    Ms3dAsciiEncoder,
    Ms3dBinEncoder,
    ObjEncoder,
    OlDecoder,
    PngEncoder,
)
from .file.data import FileData, ImageData, ModelData, TextureData
from .utils import convert, model
from .utils.convert import (
    mcsa_to_dae,
    mcsa_to_ms3d,
    mcsa_to_ms3d_ascii,
    mcsa_to_obj,
    mic_to_png,
    ol_to_dds,
)


__all__ = (
    "DaeEncoder",
    "DdsEncoder",
    "McsaDecoder",
    "MicDecoder",
    "Ms3dAsciiEncoder",
    "Ms3dBinEncoder",
    "ObjEncoder",
    "OlDecoder",
    "PngEncoder",
    "FileData",
    "ImageData",
    "ModelData",
    "TextureData",
    "convert",
    "model",
    "mcsa_to_dae",
    "mcsa_to_ms3d",
    "mcsa_to_ms3d_ascii",
    "mcsa_to_obj",
    "mic_to_png",
    "ol_to_dds",
)



from pathlib import Path

from scfile.utils.types import PathLike

from .decode import FileDecodingError, FileUnsupportedError


class OlDecodingError(FileDecodingError):
    """Base ol texture file exception."""

    def __str__(self):
        return f"Texture '{self.posix_path}'"


class OlUnknownFourcc(OlDecodingError, FileUnsupportedError):
    """Exception occurring when texture have unknown fourcc."""

    def __init__(self, path: PathLike, fourcc: str):
        self.path = Path(path)
        self.fourcc = fourcc

    def __str__(self):
        return f"{super().__str__()} has unknown format '{self.fourcc}'"



from dataclasses import dataclass, field
from typing import Dict, List

from scfile.utils.model.datatypes import Polygon, Texture, Vector


@dataclass
class Vertex:
    position: Vector = field(default_factory=Vector)
    texture: Texture = field(default_factory=Texture)
    normals: Vector = field(default_factory=Vector)
    link: Dict[int, float] = field(default_factory=dict)
    """key: bone id, value: weight"""
    # color: Color = field(default_factory=Color)


@dataclass
class Count:
    max_links: int = 0
    bones: int = 0
    vertices: int = 0
    polygons: int = 0


@dataclass
class Mesh:
    name: str = "name"
    material: str = "material"
    count: Count = field(default_factory=Count)
    vertices: List[Vertex] = field(default_factory=list)
    polygons: List[Polygon] = field(default_factory=list)
    bones: Dict[int, int] = field(default_factory=dict)
    """key: link id, value: bone id"""

    global_polygons: List[Polygon] = field(default_factory=list)
    """Empty before `convert_polygons_to_global` is called for model."""

    def resize(self) -> None:
        """Fills vertices & polygons by their counts."""
        self.vertices = [Vertex() for _ in range(self.count.vertices)]
        self.polygons = [Polygon() for _ in range(self.count.polygons)]



from abc import ABC, abstractmethod
from typing import Generic, Optional, Self, TypeVar

from scfile.enums import FileMode
from scfile.file.data import FileData
from scfile.io import BinaryBytesIO
from scfile.utils.types import PathLike

from .file import BaseFile


DATA = TypeVar("DATA", bound=FileData)


class FileEncoder(BaseFile, Generic[DATA], ABC):
    def __init__(self, data: DATA):
        self._buffer = BinaryBytesIO()
        self._data = data

    @property
    def buffer(self):
        return self._buffer

    @property
    def data(self):
        return self._data

    @property
    def b(self):
        """Bytes buffer abbreviation."""
        return self._buffer

    @property
    def magic(self) -> Optional[list[int]]:
        """Optional file magic value."""
        return None

    @property
    def content(self) -> bytes:
        """Buffer content bytes."""
        return self.b.getvalue()

    @abstractmethod
    def serialize(self) -> None:
        """Writing values from parsed data to buffer."""
        pass

    def encode(self) -> Self:
        """File encoding. Writing into buffer magic value and parsed data."""
        self.add_magic()
        self.serialize()
        return self

    def add_magic(self) -> None:
        """Writing into buffer magic value."""
        if self.magic:
            self.b.write(bytes(self.magic))

    def save_as(self, path: PathLike) -> None:
        """Saves content bytes to file at specified path. Buffer remains open."""
        with open(path, FileMode.WRITE) as fp:
            fp.write(self.content)

    def save(self, path: PathLike) -> None:
        """Saves content bytes to file at specified path. Buffer closes."""
        self.save_as(path)
        self.close()



from scfile.consts import Magic
from scfile.enums import StructFormat as F
from scfile.file.base import FileEncoder
from scfile.file.data import TextureData

from .header import DDS
from .masks import BGRA8, RGBA8, Mask


class DdsEncoder(FileEncoder[TextureData]):
    @property
    def magic(self):
        return Magic.DDS

    def serialize(self):
        self.b.writeb(F.U32, DDS.HEADER.SIZE)
        self.b.writeb(F.U32, self.flags)
        self.b.writeb(F.U32, self.data.height)
        self.b.writeb(F.U32, self.data.width)
        self.b.writeb(F.U32, self.pitch_or_linear_size)
        self.b.writen(count=1)  # Depth
        self.b.writeb(F.U32, self.data.mipmap_count)  # MipMapsCount
        self.b.writen(count=11)  # Reserved

        self._add_pixelformat()
        self._add_caps()

        self.b.write(self.data.image)

    def _add_pixelformat(self):
        self.b.writeb(F.U32, DDS.PF.SIZE)

        if self.is_compressed:
            self._add_pf_fourcc()

        else:
            self._add_pf_rgb()

    def _add_pf_fourcc(self):
        self.b.writeb(F.U32, DDS.PF.FLAG.FOURCC)
        self.b.write(self.data.fourcc)  # FourCC
        self.b.writen(5)  # BitCount & BitMasks

    def _add_pf_rgb(self):
        self.b.writeb(F.U32, DDS.PF.RGB_FLAGS)
        self.b.writen(count=1)  # FourCC
        self.b.writeb(F.U32, DDS.PF.BIT_COUNT)  # BitCount

        # BitMasks
        mask: Mask = RGBA8

        if self.data.fourcc == b"BGRA8":
            mask = BGRA8

        self.b.writeb(F.U32 * 4, *mask)

    def _add_caps(self):
        self.b.writeb(F.U32, DDS.TEXTURE | DDS.COMPLEX | DDS.MIPMAP)  # Caps1
        self.b.writeb(F.U32, self.cubemap)  # Caps2
        self.b.writen(count=3)  # Reserved

    @property
    def is_compressed(self) -> bool:
        return self.data.fourcc in (b"DXT1", b"DXT3", b"DXT5", b"ATI2")

    @property
    def flags(self) -> int:
        if self.is_compressed:
            return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.LINEARSIZE
        return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.PITCH

    @property
    def pitch(self) -> int:
        bytes_per_pixel = 4
        aligned_width = (self.data.width * bytes_per_pixel + 3) & ~3
        return aligned_width

    @property
    def pitch_or_linear_size(self) -> int:
        if self.is_compressed:
            return self.data.linear_size
        return self.pitch

    @property
    def cubemap(self) -> int:
        return 0



from pathlib import Path
from typing import Optional, TypeVar

from scfile import exceptions as exc
from scfile.enums import FileSuffix
from scfile.file.base import FileDecoder, FileEncoder
from scfile.file.data.base import FileData
from scfile.file.formats.dae import DaeEncoder
from scfile.file.formats.dds import DdsEncoder
from scfile.file.formats.mcsa import McsaDecoder
from scfile.file.formats.mic import MicDecoder
from scfile.file.formats.ms3d import Ms3dBinEncoder
from scfile.file.formats.ms3d_ascii import Ms3dAsciiEncoder
from scfile.file.formats.obj import ObjEncoder
from scfile.file.formats.ol import OlDecoder
from scfile.file.formats.png import PngEncoder
from scfile.io.binary import BinaryFileIO
from scfile.utils.types import PathLike


OPENER = TypeVar("OPENER", bound=BinaryFileIO)
DATA = TypeVar("DATA", bound=FileData)


def mcsa_to_dae(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting model `.mcsa` file to `.dae`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output `.dae` file.
        Defaults to source path with new suffix.

    Example:
        `mcsa_to_dae("model.mcsa", "model.dae")`
    """

    _convert(source, output, McsaDecoder, DaeEncoder, FileSuffix.DAE)


def mcsa_to_ms3d(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting model `.mcsa` file to `.ms3d`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output `.ms3d` file.
        Defaults to source path with new suffix.

    Example:
        `mcsa_to_ms3d("model.mcsa", "model.ms3d")`
    """

    _convert(source, output, McsaDecoder, Ms3dBinEncoder, FileSuffix.MS3D)


def mcsa_to_ms3d_ascii(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting model `.mcsa` file to `.txt`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output `.txt` file.
        Defaults to source path with new suffix.

    Example:
        `mcsa_to_ms3d_ascii("model.mcsa", "model.txt")`
    """

    _convert(source, output, McsaDecoder, Ms3dAsciiEncoder, FileSuffix.TXT)


def mcsa_to_obj(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting model `.mcsa` file to `.obj`.

    Args:
        source: Path to `.mcsa` file.
        output (optional): Path to output `.obj` file.
        Defaults to source path with new suffix.

    Example:
        `mcsa_to_obj("model.mcsa", "model.obj")`
    """

    _convert(source, output, McsaDecoder, ObjEncoder, FileSuffix.OBJ)


def mic_to_png(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting image `.mic` file to `.png`.

    Args:
        source: Path to `.mic` file.
        output (optional): Path to output `.png` file.
        Defaults to source path with new suffix.

    Example:
        `mic_to_png("image.mic", "image.png")`
    """

    _convert(source, output, MicDecoder, PngEncoder, FileSuffix.PNG)


def ol_to_dds(source: PathLike, output: Optional[PathLike] = None):
    """
    Converting texture `.ol` file to `.dds`.

    Args:
        source: Path to `.ol` file.
        output (optional): Path to output `.dds` file.
        Defaults to source path with new suffix.

    Example:
        `ol_to_dds("texture.ol", "texture.dds")`
    """

    _convert(source, output, OlDecoder, DdsEncoder, FileSuffix.DDS)


def auto(source: PathLike, output: Optional[PathLike] = None):
    """
    Automatically determines which format convert to.

    Args:
        source: Path to encrypted file.
        output (optional): Path to output decrypted file.
        Defaults to source path with new suffix.

    Raises:
        FileSuffixUnsupported - if source suffix not in consts.SUPPORTED_SUFFIXES.

    Example:
        `auto("file.mic", "file.png")`
    """

    path = Path(source)

    match path.suffix.lstrip("."):
        case FileSuffix.MCSA | FileSuffix.MCVD:
            mcsa_to_obj(source, output)
            mcsa_to_ms3d(source, output)

        case FileSuffix.MIC:
            mic_to_png(source, output)

        case FileSuffix.OL:
            ol_to_dds(source, output)

        case _:
            raise exc.FileSuffixUnsupported(path)


def _convert(
    source: PathLike,
    output: Optional[PathLike],
    decoder: type[FileDecoder[OPENER, DATA]],
    encoder: type[FileEncoder[DATA]],
    new_suffix: str,
):
    src = Path(source)
    new_src = src.with_suffix(f".{new_suffix}")

    dest = Path(output) if output else new_src

    # Check that file exists
    if not src.exists() or not src.is_file():
        raise exc.FileNotFound(src)

    # If destination is directory save file in this directory
    if dest.is_dir():
        dest = Path(dest, new_src.name)

    # Convert and save file
    with decoder(src) as dec:
        dec.convert_to(encoder).save(dest)



from scfile.consts import McsaModel
from scfile.file.base import FileEncoder
from scfile.file.data import ModelData
from scfile.utils.model import Mesh


class Ms3dAsciiEncoder(FileEncoder[ModelData]):
    def serialize(self):
        self.model = self.data.model
        self.meshes = self.data.model.meshes
        self.skeleton = self.model.skeleton
        self.flags = self.data.model.flags

        self.model.ensure_unique_names()
        self.skeleton.convert_to_local()

        self._add_header()
        self._add_meshes()
        self._add_materials()
        self._add_skeleton()
        self._add_stats()

    def _add_header(self):
        self.b.writes(
            "\n".join(
                [
                    "// MilkShape 3D ASCII",
                    "",
                    "Frames: 30",
                    "Frame: 1",
                    f"Meshes: {len(self.meshes)}",
                ]
            )
        )
        self.b.writes("\n")

    def _add_meshes(self):
        for index, mesh in enumerate(self.meshes):
            self.b.writes(f'"{mesh.name}" 0 {index}\n')
            self._add_vertices(mesh)
            self._add_normals(mesh)
            self._add_polygons(mesh)
        self.b.writes("\n")

    def _add_vertices(self, mesh: Mesh):
        self.b.writes(f"{len(mesh.vertices)}\n")
        self.b.writes(
            "\n".join(
                [
                    f"0 {v.position.x} {v.position.y} {v.position.z} {v.texture.u} {v.texture.v} {McsaModel.ROOT_BONE_ID}"
                    for v in mesh.vertices
                ]
            )
        )
        self.b.writes("\n")

    def _add_normals(self, mesh: Mesh):
        self.b.writes(f"{len(mesh.vertices)}\n")
        self.b.writes(
            "\n".join([f"{v.normals.x} {v.normals.y} {v.normals.z}" for v in mesh.vertices])
        )
        self.b.writes("\n")

    def _add_polygons(self, mesh: Mesh):
        self.b.writes(f"{len(mesh.polygons)}\n")

        for p in mesh.polygons:
            self.b.writes(f"0 {p.a} {p.b} {p.c} {p.a} {p.b} {p.c} 1\n")

    def _add_materials(self):
        self.b.writes(f"Materials: {len(self.model.meshes)}\n")

        for index, mesh in enumerate(self.model.meshes):
            blank_name = f"material_{index}"
            self.b.writes(
                "\n".join(
                    [
                        f'"{mesh.material or blank_name}"',
                        "0.200000 0.200000 0.200000 1.000000",
                        "0.800000 0.800000 0.800000 1.000000",
                        "0.000000 0.000000 0.000000 1.000000",
                        "0.000000 0.000000 0.000000 1.000000",
                        "0.000000",
                        "1.000000",
                        '""',
                        '""',
                    ]
                )
            )
            self.b.writes("\n")
        self.b.writes("\n")

    def _add_skeleton(self):
        if not self.flags.skeleton:
            self.b.writes("Bones: 0\n")
            return

        self.b.writes(f"Bones: {len(self.skeleton.bones)}\n")

        for b in self.skeleton.bones:
            self.b.writes(f'"{b.name}"\n')

            if b.parent_id < 0:
                self.b.writes('""\n')
            else:
                parent_name = self.skeleton.bones[b.parent_id].name
                self.b.writes(f'"{parent_name}"\n')

            self.b.writes(
                f"0 {b.position.x} {b.position.y} {b.position.z} {b.rotation.x} {b.rotation.y} {b.rotation.z}\n"
            )
            self.b.writes("0\n")
            self.b.writes("0\n")

    def _add_stats(self):
        self.b.writes(
            "\n".join(
                [
                    "GroupComments: 0",
                    "MaterialComments: 0",
                    "BoneComments: 0",
                    "ModelComment: 0",
                ]
            )
        )



from dataclasses import dataclass, field
from typing import Dict, List, Self

from scfile.consts import McsaModel
from scfile.utils.model.datatypes import Vector


ROOT = McsaModel.ROOT_BONE_ID


@dataclass
class Bone:
    id: int = 0
    name: str = "bone"
    parent_id: int = ROOT
    position: Vector = field(default_factory=Vector)
    rotation: Vector = field(default_factory=Vector)
    children: List[Self] = field(default_factory=list)


@dataclass
class Local:
    scale: float = 1.0
    position: Vector = field(default_factory=Vector)
    rotation: Vector = field(default_factory=Vector)


@dataclass
class Skeleton:
    bones: List[Bone] = field(default_factory=list)

    def convert_to_local(self):
        """Update bones positions by their parent bone."""

        parent_id = 0
        bones = self.bones

        for bone in bones:
            bone.rotation = Vector()
            parent_id = bone.parent_id

            while parent_id > ROOT:
                parent = bones[parent_id]
                bone.position -= parent.position
                parent_id = parent.parent_id

    def build_hierarchy(self):
        """Fills bones children list."""

        # Create a dictionary to map bone id to bones
        bone_dict: Dict[int, Bone] = {bone.id: bone for bone in self.bones}

        # Find the root bones (those without a parent)
        root_bones: List[Bone] = []

        # Assign children to their respective parents
        for bone in self.bones:
            if bone.parent_id == ROOT:
                root_bones.append(bone)
                continue

            parent_bone = bone_dict.get(bone.parent_id)
            if parent_bone:
                parent_bone.children.append(bone)

        return root_bones



from collections import defaultdict

from scfile import exceptions as exc
from scfile.consts import Factor, McsaModel, McsaSize, Signature
from scfile.enums import ByteOrder
from scfile.enums import StructFormat as F
from scfile.file.base import FileDecoder
from scfile.file.data import ModelData
from scfile.file.formats.dae import DaeEncoder
from scfile.file.formats.ms3d import Ms3dBinEncoder
from scfile.file.formats.ms3d_ascii import Ms3dAsciiEncoder
from scfile.file.formats.obj import ObjEncoder
from scfile.io import McsaFileIO
from scfile.utils.model import Bone, Mesh, Model

from .flags import Flag
from .versions import SUPPORTED_VERSIONS, VERSION_FLAGS


class McsaDecoder(FileDecoder[McsaFileIO, ModelData]):
    def to_dae(self):
        return self.convert_to(DaeEncoder)

    def to_obj(self):
        return self.convert_to(ObjEncoder)

    def to_ms3d(self):
        return self.convert_to(Ms3dBinEncoder)

    def to_ms3d_ascii(self):
        return self.convert_to(Ms3dAsciiEncoder)

    @property
    def opener(self):
        return McsaFileIO

    @property
    def order(self):
        return ByteOrder.LITTLE

    @property
    def signature(self):
        return Signature.MCSA

    def create_data(self):
        return ModelData(self.model)

    def parse(self):
        self._create_model()
        self._parse_header()
        self._parse_meshes()

        if self.flags[Flag.SKELETON]:
            self._parse_skeleton()

    def _create_model(self):
        self.model = Model()

    @property
    def scale(self):
        return self.model.scale

    @property
    def skeleton(self):
        return self.model.skeleton

    def _parse_header(self):
        self._parse_version()
        self._parse_flags()
        self._parse_scales()

    def _parse_version(self):
        self.version = self.f.readb(F.F32)

        if self.version not in SUPPORTED_VERSIONS:
            raise exc.McsaUnsupportedVersion(self.path, self.version)

    def _parse_flags(self):
        self.flags: defaultdict[int, bool] = defaultdict(bool)
        self.flags_count = VERSION_FLAGS.get(self.version)

        if not self.flags_count:
            raise exc.McsaUnsupportedVersion(self.path, self.version)

        for index in range(self.flags_count):
            self.flags[index] = self.f.readb(F.BOOL)

        self.model.flags.skeleton = self.flags[Flag.SKELETON]
        self.model.flags.texture = self.flags[Flag.TEXTURE]
        self.model.flags.normals = self.flags[Flag.NORMALS]

    def _parse_scales(self):
        self.model.scale.position = self.f.readb(F.F32)

        if self.flags[Flag.TEXTURE]:
            self.model.scale.texture = self.f.readb(F.F32)

        # ! unconfirmed
        if self.flags[Flag.NORMALS] and self.version >= 10.0:
            self.model.scale.normals = self.f.readb(F.F32)

    def _parse_meshes(self):
        self.meshes_count = self.f.readb(F.U32)

        for _ in range(self.meshes_count):
            self._parse_mesh()

    def _parse_mesh(self):
        self.mesh = Mesh()

        # Name & Material
        self.mesh.name = self.f.readstring()
        self.mesh.material = self.f.readstring()

        # Skeleton bone indexes
        if self.flags[Flag.SKELETON]:
            self._parse_bone_indexes()

        # Counts
        self.count.vertices = self.f.readcounts()
        self.count.polygons = self.f.readcounts()
        self.mesh.resize()

        # ! unknown, unconfirmed
        if self.flags[Flag.TEXTURE]:
            self.model.scale.weight = self.f.readb(F.F32)

        # ! unconfirmed
        if self.version >= 10.0:
            self._parse_locals()

        # Geometric vertices
        self._parse_position()

        # Texture vertices
        if self.flags[Flag.TEXTURE]:
            self._parse_texture()

        # ! unconfirmed
        if self.flags[Flag.BITANGENTS]:
            self._skip_vertices(size=4)

        # Vertex normals
        if self.flags[Flag.NORMALS]:
            self._parse_normals()

        # ! unconfirmed
        if self.flags[Flag.TANGENTS]:
            self._skip_vertices(size=4)

        # Vertex links
        if self.flags[Flag.SKELETON]:
            self._parse_links()

        # Vertex colors
        if self.flags[Flag.COLORS]:
            self._parse_colors()

        # Polygon faces
        self._parse_polygons()

        self.model.meshes.append(self.mesh)

    @property
    def count(self):
        return self.mesh.count

    @property
    def vertices(self):
        return self.mesh.vertices

    def _parse_bone_indexes(self):
        self.count.max_links = self.f.readb(F.U8)
        self.count.bones = self.f.readb(F.U8)

        for index in range(self.count.bones):
            self.mesh.bones[index] = self.f.readb(F.U8)

    def _parse_locals(self):
        # Possibly local axis and center (6 floats)
        # Quite useless
        self.f.read(6 * 4)

        # Another unknown float
        if self.version >= 11.0:
            self.f.read(4)

    def _parse_position(self):
        count = self.count.vertices
        scale = self.scale.position
        xyzw = self.f.readvertex(F.I16, Factor.I16 + 1, McsaSize.POSITION, count, scale)

        for vertex, (x, y, z, _) in zip(self.vertices, xyzw):
            vertex.position.x = x
            vertex.position.y = y
            vertex.position.z = z

    def _parse_texture(self):
        count = self.count.vertices
        scale = self.scale.texture
        uv = self.f.readvertex(F.I16, Factor.I16, McsaSize.TEXTURE, count, scale)

        for vertex, (u, v) in zip(self.vertices, uv):
            vertex.texture.u = u
            vertex.texture.v = v

    def _parse_normals(self):
        count = self.count.vertices
        xyzw = self.f.readvertex(F.I8, Factor.I8, McsaSize.NORMALS, count)

        for vertex, (x, y, z, _) in zip(self.vertices, xyzw):
            vertex.normals.x = x
            vertex.normals.y = y
            vertex.normals.z = z

    def _parse_links(self):
        match self.count.max_links:
            case 0:
                pass
            case 1 | 2:
                self._skip_vertices(size=4)
            case 3 | 4:
                self._skip_vertices(size=8)
            case _:
                raise exc.McsaUnknownLinkCount(self.path, self.count.max_links)

    def _parse_packed_links(self):
        linkids, linkweights = self.f.readlinkspacked(self.count.vertices, self.mesh.bones)

        for vertex, ids, weights in zip(self.vertices, linkids, linkweights):
            vertex.link = dict(zip(ids, weights))

    def _parse_plains_links(self):
        linkids, linkweights = self.f.readlinksplains(self.count.vertices, self.mesh.bones)

        for vertex, ids, weights in zip(self.vertices, linkids, linkweights):
            vertex.link = dict(zip(ids, weights))

    def _parse_colors(self):
        # Quite useless
        self._skip_vertices(size=4)

    def _parse_polygons(self):
        abc = self.f.readpolygons(self.count.polygons)

        for polygon, (a, b, c) in zip(self.mesh.polygons, abc):
            polygon.a = a
            polygon.b = b
            polygon.c = c

    def _parse_skeleton(self):
        bones_count = self.f.readb(F.U8)

        for index in range(bones_count):
            self._parse_bone(index)

    def _parse_bone(self, index: int) -> None:
        self.bone = Bone()

        self.bone.id = index
        self.bone.name = self.f.readstring()

        parent_id = self.f.readb(F.U8)
        self.bone.parent_id = parent_id if parent_id != index else McsaModel.ROOT_BONE_ID

        self._parse_bone_position()
        self._parse_bone_rotation()

        self.skeleton.bones.append(self.bone)

    def _parse_bone_position(self):
        self.bone.position.x = self.f.readb(F.F32)
        self.bone.position.y = self.f.readb(F.F32)
        self.bone.position.z = self.f.readb(F.F32)

    def _parse_bone_rotation(self):
        self.bone.rotation.x = self.f.readb(F.F32)
        self.bone.rotation.y = self.f.readb(F.F32)
        self.bone.rotation.z = self.f.readb(F.F32)

    def _skip_vertices(self, size: int = 4):
        self.f.read(self.count.vertices * size)



from .datatypes import Color, Polygon, Texture, Vector
from .mesh import Count, Mesh, Vertex
from .model import Flags, Model, Scale
from .skeleton import Bone, Local, Skeleton


__all__ = (
    "Color",
    "Polygon",
    "Texture",
    "Vector",
    "Count",
    "Mesh",
    "Vertex",
    "Flags",
    "Model",
    "Scale",
    "Bone",
    "Local",
    "Skeleton",
)



from .dae import DaeEncoder
from .dds import DdsEncoder
from .mcsa import McsaDecoder
from .mic import MicDecoder
from .ms3d import Ms3dBinEncoder
from .ms3d_ascii import Ms3dAsciiEncoder
from .obj import ObjEncoder
from .ol import OlDecoder
from .png import PngEncoder


__all__ = (
    "DaeEncoder",
    "DdsEncoder",
    "McsaDecoder",
    "MicDecoder",
    "Ms3dAsciiEncoder",
    "Ms3dBinEncoder",
    "ObjEncoder",
    "OlDecoder",
    "PngEncoder",
)



from pathlib import Path

from scfile.utils.types import PathLike

from .decode import FileDecodingError, FileParsingError, FileUnsupportedError


class McsaDecodingError(FileDecodingError):
    """Base mcsa model file exception."""

    def __str__(self):
        return f"Model '{self.posix_path}'"


class McsaCountsLimit(McsaDecodingError, FileParsingError):
    """Exception occurring when model counts is not read correctly."""

    def __init__(self, path: PathLike, counts: int):
        self.path = Path(path)
        self.counts = counts

    def __str__(self):
        return f"{super().__str__()} has invalid format."


class McsaUnknownLinkCount(McsaDecodingError, FileParsingError):
    """Exception occurring when model skeleton bones have unknown link count."""

    def __init__(self, path: PathLike, link_count: int):
        self.path = Path(path)
        self.link_count = link_count

    def __str__(self):
        return f"{super().__str__()} has unknown bones link count: {self.link_count}"


class McsaUnsupportedVersion(McsaDecodingError, FileUnsupportedError):
    """Exception occurring when model version unsupported."""

    def __init__(self, path: PathLike, version: float):
        self.path = Path(path)
        self.version = version

    def __str__(self):
        return f"{super().__str__()} has unsupported version: {self.version}"



import xml.etree.ElementTree as etree

import numpy as np

from scfile.file.base import FileEncoder
from scfile.file.data import ModelData


class DaeEncoder(FileEncoder[ModelData]):
    def serialize(self):
        self.model = self.data.model
        self.flags = self.data.model.flags
        self.skeleton = self.data.model.skeleton

        self.model.ensure_unique_names()
        # self.skeleton.convert_to_local()
        # self.skeleton.build_hierarchy()

        self._create_root()
        self._add_asset()
        self._add_geometries()
        self._add_scenes()
        self._render_xml()

    def _create_root(self):
        xmlns = "http://www.collada.org/2008/03/COLLADASchema"
        self.root = etree.Element("COLLADA", xmlns=xmlns, version="1.5.0")

    def _add_asset(self):
        asset = etree.SubElement(self.root, "asset")
        etree.SubElement(asset, "unit", name="meter", meter="1")
        etree.SubElement(asset, "up_axis").text = "Y_UP"

    def _add_geometries(self):
        library = etree.SubElement(self.root, "library_geometries")

        for mesh in self.model.meshes:
            geom = etree.SubElement(library, "geometry", id=f"{mesh.name}-mesh", name=mesh.name)
            self.node = etree.SubElement(geom, "mesh")

            self.mesh = mesh

            self._add_positions()
            self._add_normals()
            self._add_texture()
            self._add_vertices()
            self._add_triangles()

    def _add_positions(self):
        data = np.array([(v.position.x, v.position.y, v.position.z) for v in self.mesh.vertices])
        self._add_source("positions", data, ["X", "Y", "Z"])

    def _add_normals(self):
        data = np.array([(v.normals.x, v.normals.y, v.normals.z) for v in self.mesh.vertices])
        self._add_source("normals", data, ["X", "Y", "Z"])

    def _add_texture(self):
        data = np.array([(v.texture.u, -v.texture.v) for v in self.mesh.vertices])
        self._add_source("texture", data, ["S", "T"])

    def _add_source(self, name: str, data: np.ndarray, components: list[str]):
        self.source = etree.SubElement(self.node, "source", id=f"{self.mesh.name}-{name}")

        array = etree.SubElement(
            self.source,
            "float_array",
            id=f"{self.mesh.name}-{name}-array",
            count=str(self.mesh.count.vertices),
        )

        array.text = " ".join(map(str, data.flatten()))

        self._add_source_common(name, components)

    def _add_source_common(self, name: str, components: list[str]):
        common = etree.SubElement(
            self.source,
            "technique_common",
        )
        accessor = etree.SubElement(
            common,
            "accessor",
            source=f"#{self.mesh.name}-{name}-array",
            count=str(self.mesh.count.vertices),
            stride=str(len(components)),
        )

        for name in components:
            accessor.append(etree.Element("param", name=name, type="float"))

    def _add_vertices(self):
        vertices = etree.SubElement(self.node, "vertices", id=f"{self.mesh.name}-vertices")
        etree.SubElement(
            vertices, "input", semantic="POSITION", source=f"#{self.mesh.name}-positions"
        )

    def _add_triangles(self):
        self.triangles = etree.SubElement(
            self.node, "triangles", count=str(self.mesh.count.polygons)
        )
        self._add_inputs()
        self._add_polygons()

    def _add_inputs(self):
        self._add_input("VERTEX", "vertices")

        if self.flags.normals:
            self._add_input("NORMAL", "normals")

        if self.flags.texture:
            self._add_input("TEXCOORD", "texture")

    def _add_input(self, semantic: str, name: str):
        etree.SubElement(
            self.triangles,
            "input",
            semantic=semantic,
            source=f"#{self.mesh.name}-{name}",
            offset="0",
        )

    def _add_polygons(self):
        indices = np.array([vertex_id for polygon in self.mesh.polygons for vertex_id in polygon])

        p = etree.SubElement(self.triangles, "p")
        p.text = " ".join(map(str, indices))

    def _add_scenes(self):
        library = etree.SubElement(self.root, "library_visual_scenes")
        visual_scene = etree.SubElement(library, "visual_scene", id="scene")

        for mesh in self.model.meshes:
            node = etree.SubElement(
                visual_scene, "node", id=f"{mesh.name}-node", name=mesh.name, type="NODE"
            )
            etree.SubElement(node, "instance_geometry", url=f"#{mesh.name}-mesh", name=mesh.name)

        scene = etree.SubElement(self.root, "scene")
        etree.SubElement(scene, "instance_visual_scene", url="#scene")

    def _render_xml(self):
        etree.indent(self.root)
        self.buffer.write(etree.tostring(self.root))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scfile',
 'scfile.cli',
 'scfile.exceptions',
 'scfile.file',
 'scfile.file.base',
 'scfile.file.data',
 'scfile.file.formats',
 'scfile.file.formats.dae',
 'scfile.file.formats.dds',
 'scfile.file.formats.mcsa',
 'scfile.file.formats.mic',
 'scfile.file.formats.ms3d',
 'scfile.file.formats.ms3d_ascii',
 'scfile.file.formats.obj',
 'scfile.file.formats.ol',
 'scfile.file.formats.png',
 'scfile.io',
 'scfile.utils',
 'scfile.utils.model']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.7,<9.0.0',
 'lz4>=4.3.2,<5.0.0',
 'numpy>=1.26.3,<2.0.0',
 'rich>=13.0.0,<14.0.0']

entry_points = \
{'console_scripts': ['scfile = scfile.__main__:main']}

setup_kwargs = {
    'name': 'sc-file',
    'version': '3.5.4',
    'description': 'Utility & Library for decoding stalcraft assets',
    'long_description': '# SC FILE\n\nUtility and Library for decoding and converting stalcraft assets files, such as models and textures into well-known formats.\n\nDesigned for artworks creation and the like.\n\n> [!NOTE]\n> There is not and will not be encoding back into game formats.\n\n> [!WARNING]\n> Do not use game assets directly. \\\n> Any changes in game client can be detected.\n\nYou can use executable program from [Releases](https://github.com/onejeuu/sc-file/releases) page.\n\n# 📁 Formats\n\n| Type    | Source        | Output                 |\n| ------- | ------------- | ---------------------- |\n| Model   | .mcsa / .mcvd | .dae, .obj, ms3d, .txt |\n| Texture | .ol           | .dds                   |\n| Image   | .mic          | .png                   |\n\n### Models\n\n- Versions supported: 7.0, 8.0, 10.0, 11.0\n- Skeleton and Animations currently unsupported\n\n### Textures\n\n- Formats supported: DXT1, DXT3, DXT5, RGBA8, BGRA8, DXN_XY\n- Formats unsupported: RGBA32F, Cubemaps\n\n# 💻 CLI Utility\n\n## Usage\n\nFrom bash:\n\n```bash\nscfile [FILES]... [OPTIONS]\n```\n\n> [!TIP]\n> You can just drag and drop one or multiple files onto `scfile.exe`.\n\n## Arguments\n\n- `FILES`: **List of file paths to be converted**. Multiple files should be separated by **spaces**. Accepts both full and relative paths. **Does not accept directory**.\n\n## Options\n\n- `-O`, `--output`: **One path to output directory**. If not specified, file will be saved in same directory with a new suffix.\n\n## Examples\n\n1. Convert a single file:\n\n   ```bash\n   scfile file.mcsa\n   ```\n\n   _Will be saved in same directory with a new suffix._\n\n2. Convert multiple files to a specified directory:\n\n   ```bash\n   scfile file1.mcsa file2.mcsa --output path/to/dir\n   ```\n\n3. Convert all `.mcsa` files in current directory:\n\n   ```bash\n   scfile *.mcsa\n   ```\n\n   _Subdirectories are not included._\n\n4. Convert all `.mcsa` files with subdirectories to a specified directory:\n\n   ```bash\n   scfile **/*.mcsa -O path/to/dir\n   ```\n\n   _With `--output` specified, directory structure is not duplicated._\n\n# 📚 Library\n\n## Install\n\n### Pip\n\n```bash\npip install sc-file -U\n```\n\n### Manual\n\n```bash\ngit clone git@github.com:onejeuu/sc-file.git\n```\n\n```bash\ncd sc-file\n```\n\n```bash\npoetry install\n```\n\n## Usage\n\n### Simple\n\n```python\nfrom scfile import convert\n\n# Output path is optional.\n# Defaults to source path with new suffix.\nconvert.mcsa_to_obj("path/to/model.mcsa", "path/to/model.obj")\nconvert.ol_to_dds("path/to/texture.ol", "path/to/texture.dds")\nconvert.mic_to_png("path/to/image.mic", "path/to/image.png")\n\n# Skeleton support via MilkShape3D\nconvert.mcsa_to_ms3d("path/to/model.mcsa", "path/to/model.ms3d")\nconvert.mcsa_to_ms3d_ascii("path/to/model.mcsa", "path/to/model.txt")\n\n# Or determinate it automatically\nconvert.auto("path/to/model.mcsa")\n```\n\n### Advanced\n\n- Default\n\n```python\nfrom scfile.file.data import ModelData\nfrom scfile.file import McsaDecoder, ObjEncoder\n\nmcsa = McsaDecoder("model.mcsa")\ndata: ModelData = mcsa.decode()\nmcsa.close() # ? Necessary to close\n\nobj = ObjEncoder(data)\nobj.encode().save("model.obj") # ? Encoder closes after saving\n```\n\n- Use encoded content bytes\n\n```python\nobj = ObjEncoder(data)\nobj.encode()\n\nwith open("model.obj", "wb") as fp:\n    fp.write(obj.content)\n\nobj.close() # ? Necessary to close\n```\n\n- Use convert methods\n\n```python\nmcsa = McsaDecoder("model.mcsa")\nmcsa.convert_to(ObjEncoder).save("model.obj")\nmcsa.close() # ? Necessary to close\n```\n\n```python\nmcsa = McsaDecoder("model.mcsa")\nmcsa.to_obj().save("model.obj")\nmcsa.close() # ? Necessary to close\n```\n\n- Use context manager\n\n```python\nwith McsaDecoder("model.mcsa") as mcsa:\n    data: ModelData = mcsa.decode()\n\nwith ObjEncoder(data) as obj:\n    obj.encode().save("model.obj")\n```\n\n- Use context manager + convert methods\n\n```python\nwith McsaDecoder("model.mcsa") as mcsa:\n    obj = mcsa.convert_to(ObjEncoder)\n    obj.close()\n```\n\n```python\nwith McsaDecoder("model.mcsa") as mcsa:\n    mcsa.to_obj().save("model.obj")\n```\n\n> [!IMPORTANT]\n> When using `convert_to` buffer remains open. \\\n> `close()` or `save()` or another context (`with`) is necessary.\n\n- Save multiple copies\n\n```python\nwith McsaDecoder("model.mcsa") as mcsa:\n    with mcsa.to_obj() as obj:\n        obj.save_as("model_1.obj")\n        obj.save_as("model_2.obj")\n```\n\n# 🛠️ Build\n\n> [!IMPORTANT]\n> You will need [poetry](https://python-poetry.org) to do compilation.\n\n> [!TIP]\n> Recommended to create virtual environment.\n>\n> ```bash\n> poetry shell\n> ```\n\nThen install dependencies:\n\n```bash\npoetry install\n```\n\nAnd run script to compile:\n\n```bash\npython scripts/build.py\n```\n\nExecutable file will be created in `/dist` directory.\n',
    'author': 'onejeuu',
    'author_email': 'mail@66rk.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/onejeuu/sc-file',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<3.13',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Self


@dataclass
class Vector:
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __iter__(self):
        return iter((self.x, self.y, self.z))

    def __add__(self, vec: Self):
        return Vector(self.x + vec.x, self.y + vec.y, self.z + vec.z)

    def __sub__(self, vec: Self):
        return Vector(self.x - vec.x, self.y - vec.y, self.z - vec.z)


@dataclass
class Texture:
    u: float = 0.0
    v: float = 0.0

    def __iter__(self):
        return iter((self.u, self.v))


@dataclass
class Polygon:
    a: int = 0
    b: int = 0
    c: int = 0

    def __iter__(self):
        return iter((self.a, self.b, self.c))

    def __lshift__(self, offset: int):
        return Polygon(self.a - offset, self.b - offset, self.c - offset)

    def __rshift__(self, offset: int):
        return Polygon(self.a + offset, self.b + offset, self.c + offset)


@dataclass
class Color:
    r: float = 0.8
    g: float = 0.8
    b: float = 0.8

    def __iter__(self):
        return iter((self.r, self.g, self.b))



import pathlib

import click


class Types:
    FILES = click.Path(path_type=pathlib.Path, exists=True, readable=True, dir_okay=False)
    OUTPUT = click.Path(
        path_type=pathlib.Path, exists=True, writable=True, dir_okay=True, file_okay=False
    )



from scfile.file.base import FileEncoder
from scfile.file.data import ModelData
from scfile.utils.model import Mesh, Polygon


class ObjEncoder(FileEncoder[ModelData]):
    def serialize(self):
        self.model = self.data.model
        self.flags = self.data.model.flags

        self.model.ensure_unique_names()
        self.model.convert_polygons_to_global(start=1)

        for mesh in self.model.meshes:
            self.b.writes(f"o {mesh.name}\n")
            self.b.writes(f"usemtl {mesh.material}\n")

            self._add_geometric_vertices(mesh)

            if self.flags.texture:
                self._add_texture_coordinates(mesh)

            if self.flags.normals:
                self._add_vertex_normals(mesh)

            self.b.writes(f"g {mesh.name}\n")
            self._add_polygonal_faces(mesh)

    def _add_geometric_vertices(self, mesh: Mesh):
        self._write_vertex_data(
            [f"v {v.position.x} {v.position.y} {v.position.z}" for v in mesh.vertices]
        )

    def _add_texture_coordinates(self, mesh: Mesh):
        self._write_vertex_data([f"vt {v.texture.u} {1.0 - v.texture.v}" for v in mesh.vertices])

    def _add_vertex_normals(self, mesh: Mesh):
        self._write_vertex_data(
            [f"vn {v.normals.x} {v.normals.y} {v.normals.z}" for v in mesh.vertices]
        )

    def _add_polygonal_faces(self, mesh: Mesh):
        self._write_vertex_data([f"f {self._polygon_to_faces(p)}" for p in mesh.global_polygons])

    def _write_vertex_data(self, data: list[str]):
        self.b.writes("\n".join(data))
        self.b.write(b"\n\n")

    def _polygon_to_faces(self, polygon: Polygon):
        a, b, c = polygon.a, polygon.b, polygon.c

        if self.flags.texture and self.flags.normals:
            return f"{a}/{a}/{a} {b}/{b}/{b} {c}/{c}/{c}"

        if self.flags.texture:
            return f"{a}/{a} {b}/{b} {c}/{c}"

        if self.flags.normals:
            return f"{a}//{a} {b}//{b} {c}//{c}"

        return f"{a} {b} {c}"



from enum import IntFlag


class HeaderFlag(IntFlag):
    CAPS = 0x1
    HEIGHT = 0x2
    WIDTH = 0x4
    PITCH = 0x8
    PIXELFORMAT = 0x1000
    MIPMAPCOUNT = 0x20000
    LINEARSIZE = 0x80000
    DEPTH = 0x800000


class Header:
    SIZE = 124
    FLAG = HeaderFlag
    FLAGS = FLAG.CAPS | FLAG.HEIGHT | FLAG.WIDTH | FLAG.PIXELFORMAT | FLAG.MIPMAPCOUNT


class PixelFormatFlag(IntFlag):
    ALPHAPIXELS = 0x1
    ALPHA = 0x2
    FOURCC = 0x4
    RGB = 0x40


class PixelFormat:
    SIZE = 32
    BIT_COUNT = 32
    FLAG = PixelFormatFlag
    RGB_FLAGS = FLAG.RGB | FLAG.ALPHAPIXELS


class CubemapFlag(IntFlag):
    POSITIVE_X = POSX = 0x400
    NEGATIVE_X = NEGX = 0x800
    POSITIVE_Y = POSY = 0x1000
    NEGATIVE_Y = NEGY = 0x2000
    POSITIVE_Z = POSZ = 0x4000
    NEGATIVE_Z = NEGZ = 0x8000


class DDS:
    """DDS Header Structure."""

    HEADER = Header
    PF = PixelFormat
    CUBEMAP = 0x200
    CF = CubemapFlag
    CUBEMAPS = CUBEMAP | CF.POSX | CF.NEGX | CF.POSY | CF.NEGY | CF.POSZ | CF.NEGZ
    COMPLEX = 0x8
    TEXTURE = 0x1000
    MIPMAP = 0x400000
